@@ -57,13 +57,13 @@ async def handle(bot: Bot, event: Event, state: T_State, matcher: Matcher):
     else:
         got = None
 
-    if len(tiles) < 3:
+    if len(furo) == 0 and len(tiles) < 3:
         # 少于三张牌不进行计算
         return
 
     result = shanten(tiles, furo)
     with StringIO() as sio:
-        if result.shanten == -1:
+        if result.shanten == -1 and len(result.hand.furo) * 3 + len(tiles) == 14:
             # 分析和牌
             hora_ron = build_hora_from_shanten_result(
                 result, tiles[-1], False,
